#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <exec/io.h>

#include <sane/sanei_usb.h>
#include <sane/sanei_config.h>

#include <DriverIo.h>
#include <proto/poseidon.h>
#include <proto/exec.h>
#include <exec/lists.h>

//#define DEBUG

#ifdef DEBUG
#define DBG(x)  x
#else
#define DBG(x)
#endif


#define GetHead(_l)  \
({ struct List *l = (struct List *)(_l);  \
	l->lh_Head->ln_Succ ? l->lh_Head : (struct Node *)0;  \
})



static SANE_Bool inited = SANE_FALSE;
static int libusb_timeout = 30 * 100;        /* 30 seconds */

/// enum libusb_class_code{}
/*
enum libusb_class_code {
USB_CLASS_PER_INTERFACE = 0,
USB_CLASS_AUDIO = 1,
USB_CLASS_COMM = 2,
USB_CLASS_HID = 3,
USB_CLASS_PRINTER = 7,
USB_CLASS_PTP = 6,
USB_CLASS_MASS_STORAGE = 8,
USB_CLASS_HUB = 9,
USB_CLASS_DATA = 10,
USB_CLASS_VENDOR_SPEC = 0xff,
};
 */
///

#define INEP 				0		// incomming pipe
#define OUTEP 				1		// outgoing pipe
#define INTEP 				2		// interrupt pipe

#define NAK_TIMEOUT 		30000    //ms

typedef struct _Endpoint
{
    struct PsdEndpoint 	*ep_Bulk;
    struct PsdPipe 		*ep_BulkPipe;
    ULONG 			ep_BulkPktSize;
} Endpoint;

typedef struct
{
    SANE_Bool open;
    int fd;
    SANE_String devname;
    SANE_Int vendor;
    SANE_Int product;

    // poseidon stuff
    struct PsdAppBinding *psd_Pab;
    struct PsdDevice *psd_Device;
    struct PsdConfig *psd_Config;
    struct PsdInterface *psd_Interface;

    struct PsdPipe *psd_DefaultPipe;
    struct _Endpoint psd_Bulk[3];

    struct MsgPort  *psd_MsgPort;
}
device_list_type;


void internal_close(device_list_type *dev);

#ifdef __MORPHOS__
void disconnect_hook(void)
{
    device_list_type *dev = (device_list_type *) REG_A1;

    DBG(KPrintF("sanei_usb_close: disconnect_hook %x start\n", dev));

    internal_close(dev);
	
    DBG(KPrintF("sanei_usb_close: disconnect_hook end\n", dev));
}

struct EmulLibEntry hooczekConstructor =
{
    TRAP_LIB, 0, (void (*)(void)) disconnect_hook
};

struct Hook a_hook = {{NULL, NULL},
    (APTR)&hooczekConstructor,
    NULL, NULL
};

#elif __AROS__
#define M_HOOK_h(name)\
    extern struct Hook name##_hook;

#define M_HOOK(n, y, z) \
        AROS_UFP3(IPTR, n##_func, AROS_UFPA(struct Hook *, h, A0), AROS_UFPA(y, , A2), \
    AROS_UFPA(z, , A1)); \
    struct Hook n##_hook = { {NULL, NULL}, (HOOKFUNC)n##_func, NULL, NULL}; \
        AROS_UFH3(IPTR, n##_func, AROS_UFHA(struct Hook *, h, A0), AROS_UFHA(y, , A2), \
        AROS_UFHA(z, , A1)) { AROS_USERFUNC_INIT

#define M_HOOK_END AROS_USERFUNC_EXIT }

M_HOOK_h(a)

M_HOOK(a, IPTR mempool, IPTR dana)
{

}

M_HOOK_END
#endif



/* total number of devices that can be opened at the same time */
#define MAX_DEVICES 100

/* per-device information, using the functions' parameters dn as index */
static device_list_type devices[MAX_DEVICES];
/// InitDevice()
static BOOL InitDevice(device_list_type *psd)
{
    BOOL error = FALSE;
    struct List *configlist;
    struct List *interfacelist;
    struct List *altinterfacelist;
    struct List *endpointlist;
    struct PsdEndpoint *endpoint;
    Endpoint *ep;
    ULONG ifnum;
    ULONG numeps;
    ULONG altifnum;
    ULONG i;

    // getting config list
    psdGetAttrs(PGA_DEVICE, psd->psd_Device, DA_ConfigList, &configlist, TAG_END);

    psd->psd_Config = (struct PsdConfig *) GetHead(configlist);

    // getting intefaces list
    psdGetAttrs(PGA_CONFIG, psd->psd_Config, CA_InterfaceList, &interfacelist, TAG_END);

    psd->psd_Interface = (struct PsdInterface *) GetHead(interfacelist);

    psdGetAttrs(
        PGA_INTERFACE, psd->psd_Interface,
        IFA_InterfaceNum, &ifnum,
        IFA_AlternateNum, &altifnum,
        IFA_AlternateIfList, &altinterfacelist,
        IFA_EndpointList, &endpointlist,
        TAG_END);

    if(!(psd->psd_MsgPort = CreateMsgPort()))
    {
        return FALSE;
    }

    // control pipe !!!
    if((psd->psd_DefaultPipe = psdAllocPipe(psd->psd_Device, psd->psd_MsgPort, NULL)))
    {
        if(altifnum != 0)
        {
            psdSetAltInterface(psd->psd_DefaultPipe, altinterfacelist->lh_Head);
        }

        psdGetAttrs(PGA_CONFIG, psd->psd_Config, CA_InterfaceList, &interfacelist, TAG_END);

        // Get the interface and the endpoint list, then get endpoints
        psd->psd_Interface = (struct PsdInterface *) GetHead(interfacelist);
        psdGetAttrs(
            PGA_INTERFACE, psd->psd_Interface,
            IFA_InterfaceNum, &ifnum,
            IFA_NumEndpoints, &numeps,
            IFA_AlternateNum, &altifnum,
            IFA_EndpointList, &endpointlist,
            TAG_END);

        endpoint	= (struct PsdEndpoint *) GetHead(endpointlist);

        while(((struct Node *) endpoint)->ln_Succ)
        {
            ULONG epnum;
            ULONG epmaxpktsize;
            ULONG eptranstype;
            ULONG episin;

            psdGetAttrs(
                PGA_ENDPOINT, endpoint,
                EA_EndpointNum, &epnum,
                EA_TransferType, &eptranstype,
                EA_MaxPktSize, &epmaxpktsize,
                EA_IsIn, &episin,
                TAG_END);

            ep = NULL;
            switch(eptranstype)
            {
            case USEAF_BULK:
                ep = (episin == TRUE) ? &psd->psd_Bulk[INEP] : &psd->psd_Bulk[OUTEP];
                break;
            case USEAF_INTERRUPT:
                ep = (episin == TRUE) ? &psd->psd_Bulk[INTEP] : NULL;
                break;

            };
            if(ep)
            {
                if((eptranstype == USEAF_BULK || eptranstype == USEAF_INTERRUPT))
                {
                    ep->ep_Bulk = endpoint;
                    ep->ep_BulkPktSize = epmaxpktsize;

                    if((ep->ep_BulkPipe = psdAllocPipe(psd->psd_Device, psd->psd_MsgPort, ep->ep_Bulk)))
                    {
                        psdSetAttrs(
                            PGA_PIPE, ep->ep_BulkPipe,
                            PPA_AllowRuntPackets, TRUE,
                            PPA_NoShortPackets, TRUE,
                            PPA_NakTimeout, TRUE,
                            PPA_NakTimeoutTime,  NAK_TIMEOUT,
                            TAG_END);
                    }
                    else
                    {
                        error = TRUE;
                        break;
                    }
                }
            }
            else
            {
                error = TRUE;
                break;
            }

            endpoint = (struct PsdEndpoint *)((struct Node *) endpoint)->ln_Succ;
        }

        if(!error)
        {
            return TRUE;
        }
		else
		{
			for(i = 0; i < numeps; i++)
			{
				ep = &psd->psd_Bulk[i];

				if(ep && ep->ep_BulkPipe)
				{
					psdFreePipe(ep->ep_BulkPipe);
				}
			}
			psdFreePipe(psd->psd_DefaultPipe);
		}
    }

    DeleteMsgPort(psd->psd_MsgPort);
    return FALSE;
}
///
/// sanei_usb_open()

SANE_Status sanei_usb_open(const char* devname, int* dn)
{
    int devcount;
    SANE_Bool found = SANE_FALSE;
    struct PsdDevice *pd = NULL;

    DBG(KPrintF("sanei_usb_open: trying to open device `%s'\n", devname));
    if(!dn)
    {
        DBG(KPrintF("sanei_usb_open: can't open `%s': dn == NULL\n", devname));
        return SANE_STATUS_INVAL;
    }

    // traversing devices table
    for(devcount = 0; devcount < MAX_DEVICES && devices[devcount].devname != 0; devcount++)
    {
        // if we found device with proper name
        if(strcmp(devices[devcount].devname, devname) == 0)
        {
            // ... checking if this device is opened
            if(devices[devcount].open)
            {
                DBG(KPrintF("sanei_usb_open: device `%s' already open\n", devname));
                return SANE_STATUS_INVAL;
            }
            found = SANE_TRUE;
            break;
        }
    }

    if(!found)
    {
        DBG(KPrintF("sanei_usb_open: can't find device `%s' in list\n", devname));
        return SANE_STATUS_INVAL;
    }

    // getting the device
    if(!(pd = psdFindDevice(pd,
                            DA_VendorID,  devices[devcount].vendor,
                            DA_ProductID, devices[devcount].product,
                            TAG_END)))
    {
        DBG(KPrintF("Scanner %s not found\n", devname));
        return SANE_STATUS_INVAL;;
    }

    devices[devcount].psd_Device = pd;
    devices[devcount].fd = devcount;
    *dn = devcount;

    DBG(KPrintF("App binding\n"));

    if((devices[devcount].psd_Pab =
                psdClaimAppBinding(ABA_Device, devices[devcount].psd_Device,
                                   ABA_UserData, & (devices[devcount]),
                                   ABA_ReleaseHook, &a_hook,
                                   ABA_ForceRelease, TRUE,
                                   TAG_END)))
    {
		DBG(KPrintF("InitDevice\n"));
        if(InitDevice(& (devices[devcount])))
        {
            DBG(KPrintF("InitPoseidon()"));
            devices[devcount].open = TRUE;
            return SANE_STATUS_GOOD;
        }
        else
        {
            DBG(KPrintF("InitDevice() failed\n"));
            psdReleaseAppBinding(devices[devcount].psd_Pab);
			devices[devcount ].psd_Pab = NULL;
        }
    }
    DBG(KPrintF("open failed\n"));
    return SANE_STATUS_INVAL;
}

///
/// sanei_usb_close()

void internal_close(device_list_type *dev)
{
    int i, numeps;

    if(!dev->open)
    {
        return;
    }

    psdGetAttrs(PGA_INTERFACE, dev->psd_Interface,
                IFA_NumEndpoints, &numeps, TAG_END);
    for(i = 0; i < numeps; i++)
    {
        Endpoint *ep = & (dev->psd_Bulk[i]);

        if(ep->ep_BulkPipe)
        {
            psdFreePipe(ep->ep_BulkPipe);
			ep->ep_BulkPipe = NULL;
        }
    }

    psdFreePipe(dev->psd_DefaultPipe);
    dev->psd_DefaultPipe = NULL;
	
    if(dev->psd_MsgPort)
    {
        DeleteMsgPort(dev->psd_MsgPort);
		dev->psd_MsgPort = NULL;
    }

    dev->open = SANE_FALSE;
	dev->psd_Pab = NULL;
	
    return;
}


/// sanei_usb_close()

void sanei_usb_close(SANE_Int dn)
{
    ULONG i, numeps;
    DBG(KPrintF("sanei_usb_close: closing device %d\n", dn));
    if(dn >= MAX_DEVICES || dn < 0)
    {
        //DBG (1, "sanei_usb_close: dn >= MAX_DEVICES || dn < 0\n", "a")
        return;
    }
    if(!devices[dn].open)
    {
        //    DBG (1, "sanei_usb_close: device %d already closed or never opened\n", dn);
        return;
    }

    DBG(KPrintF("get iface\n"));
    psdGetAttrs(PGA_INTERFACE, devices[dn].psd_Interface,
                IFA_NumEndpoints, &numeps, TAG_END);
    for(i = 0; i < numeps; i++)
    {
        Endpoint *ep = & (devices[dn].psd_Bulk[i]);

        if(ep->ep_BulkPipe)
        {
            psdFreePipe(ep->ep_BulkPipe);
			ep->ep_BulkPipe = NULL;
        }
    }

    psdFreePipe(devices[dn].psd_DefaultPipe);
	devices[dn].psd_DefaultPipe = 0;

    if(devices[dn].psd_MsgPort)
    {
        DeleteMsgPort(devices[dn].psd_MsgPort);
		devices[dn].psd_MsgPort = NULL;
    }

    devices[dn].open = SANE_FALSE;

	if (devices[dn].psd_Pab)
	{
		psdReleaseAppBinding(devices[dn].psd_Pab);
		devices[dn].psd_Pab = NULL;
	}
    return;
}


///
/// sanei_usb_read_bulk()

int sanei_usb_read_bulk(int dn, char* buf, int *size)
{
    if(!size)
    {
        DBG(KPrintF("sanei_usb_read_bulk: size == NULL\n"));
        return SANE_STATUS_INVAL;
    }

    if((dn >= MAX_DEVICES) || (dn < 0) || !devices[dn].open)
    {
        DBG(KPrintF("sanei_usb_read_bulk: dn >= MAX_DEVICES || dn < 0\n"));
        return SANE_STATUS_INVAL;
    }
    else
    {
        struct PsdPipe *inep =  devices[dn].psd_Bulk[INEP].ep_BulkPipe;
        LONG ioerr, len;

        DBG(KPrintF("sanei_usb_read_bulk: trying to read %lu bytes\n", (unsigned long) *size));

        ioerr = psdDoPipe(inep, buf, *size);
        len = psdGetPipeActual(inep);
        if(!len)
        {
            ioerr = psdDoPipe(inep, buf, *size);
            len = psdGetPipeActual(inep);
        }
        *size = len;
        if(len > 0)
        {
            DBG(KPrintF("sanei_usb_read_bulk: read %lu bytes correctly\n", (unsigned long) *size));
            return SANE_STATUS_GOOD;
        }
        else if(!ioerr)
        {
            DBG(KPrintF("sanei_usb_read_bulk: EOF\n"));
            return SANE_STATUS_EOF;

        }

        DBG(KPrintF("sanei_usb_read_bulk(): eof", psdNumToStr(NTS_IOERR, ioerr, "Unknown Error"), ioerr));
        *size = 0;
    }
    return SANE_STATUS_IO_ERROR;
}

///
/// sanei_usb_write_bulk()

int sanei_usb_write_bulk(int dn, char* buf, int *size)
{
    LONG ioerr, len;

    if(!size)
    {
        DBG(KPrintF("sanei_usb_write_bulk: size == NULL\n"));
        return SANE_STATUS_INVAL;
    }

    if((dn >= MAX_DEVICES) || (dn < 0) || !devices[dn].open)
    {
        DBG(KPrintF("sanei_usb_write_bulk: dn >= MAX_DEVICES || dn < 0\n"));
        return SANE_STATUS_INVAL;
    }
    else
    {
        struct PsdPipe *outep =  devices[dn].psd_Bulk[OUTEP].ep_BulkPipe;
        DBG(KPrintF("sanei_usb_write_bulk: trying to write %lu bytes\n", (unsigned long) *size));

        if(!(ioerr = psdDoPipe(outep, buf, *size)))
        {
            return SANE_STATUS_GOOD;
        }

        DBG(KPrintF("sanei_usb_write_bulk() %s (ioerr %ld)\n", psdNumToStr(NTS_IOERR, ioerr, "Unknown Error"), ioerr));
    }
    return SANE_STATUS_IO_ERROR;

}

///
/// sanei_usb_get_vendor_product()  - done
// dn - number of table element to get vendor and product id from
int sanei_usb_get_vendor_product(int dn,  SANE_Word *vendor, SANE_Word *product)
{
    SANE_Word vendorID = 0;
    SANE_Word productID = 0;

    if((dn >= MAX_DEVICES) || (dn < 0))
    {
        return SANE_STATUS_INVAL;
    }

    vendorID = devices[dn].vendor;
    productID = devices[dn].product;

    if(vendor)
    {
        *vendor = vendorID;
    }
    if(product)
    {
        *product = productID;
    }

    if(!vendorID || !productID)
    {
        return SANE_STATUS_UNSUPPORTED;
    }

    return SANE_STATUS_GOOD;
}

///
/// sanei_usb_init() - done

int sanei_usb_init(void)
{
    struct PsdDevice *pd = NULL;
    APTR pp;
    struct MsgPort *mp;
    LONG devvendorid, devprodid, devnumcfgs;
    STRPTR devprodname;
    struct List *cfgs; // configurations
    struct Node *pc;
    struct List *ifs; // interfaces
    struct Node *pif, *altpif;
    LONG ifclass;
    int found = FALSE;
    SANE_Int dn = 0;
    char devname[1000];

    //if was inited earlier
    if(inited)
    {
        return 1;
    }


    // zeroing table of connected devices
    memset(devices, 0, sizeof(devices));

    pd = NULL;
    mp = CreateMsgPort();			// creating communication port
    psdLockReadPBase();				// locking poseidon database

    DBG(KPrintF("sanei_usb_init device loop\n"));

    // all devices loop
    while(pd = psdGetNextDevice(pd))
    {
        psdLockReadDevice(pd);		// getting information
        psdGetAttrs(PGA_DEVICE, pd,
                    DA_VendorID, &devvendorid,
                    DA_ProductID, &devprodid,
                    DA_NumConfigs, &devnumcfgs,
                    DA_ConfigList, &cfgs,
                    DA_ProductName, &devprodname,
                    TAG_END);

        DBG(KPrintF("sanei_usb_init: device 0x%04x/0x%04x\n", devvendorid, devnumcfgs));
        if(!devnumcfgs)
        {
            DBG(KPrintF("sanei_usb_init: device 0x%04x/0x%04x is not configured\n", devvendorid, devnumcfgs));
			psdUnlockDevice(pd);
            continue;
        }
        if(devvendorid == 0 || devprodid == 0)
        {
            DBG(KPrintF("sanei_usb_init: device 0x%04x/0x%04x looks like a root hub\n", devvendorid, devnumcfgs));
			psdUnlockDevice(pd);
            continue;
        }

        // configuration loop
        pc = cfgs->lh_Head;
        while(pc->ln_Succ)
        {
            DBG(KPrintF("sanei_usb_init configuration loop\n"));
            psdGetAttrs(PGA_CONFIG, pc, CA_InterfaceList, &ifs, TAG_END);
            pif = ifs->lh_Head;

            // interfaces loop
            while(pif->ln_Succ)
            {
                altpif = pif;

                psdGetAttrs(PGA_INTERFACE, altpif, IFA_Class, &ifclass, TAG_END);

                DBG(KPrintF("sanei_usb_init: class test: %d\n", ifclass));
                switch(ifclass)
                {

                case USB_CLASS_VENDOR_SPEC:
                case USB_CLASS_PER_INTERFACE:
                case 16:
                    // case USB_CLASS_HID:
                    DBG(KPrintF("Scanner? found.\n"));
                    found = SANE_TRUE;

                    // inserting scanner data into device table
                    snprintf(devname, sizeof(devname), "poseidon:%d:%d", devvendorid, devprodid);
                    devices[dn].devname = strdup(devname);

                    if(!devices[dn].devname)
                    {
                        psdUnlockDevice(pd);
                        psdUnlockPBase();
                        DeleteMsgPort(mp);

                        return 0;
                    }

                    DBG(KPrintF("%s\n", devname));

                    devices[dn].vendor = devvendorid;
                    devices[dn].product = devprodid;
                    devices[dn].open = SANE_FALSE;
                    devices[dn].fd = 1;
                    //devices[dn].psd_Interface = alpif;

                    dn++;
                    // too many usb devices
                    if(dn >= MAX_DEVICES)
                    {
                        psdUnlockDevice(pd);
                        psdUnlockPBase();
                        DeleteMsgPort(mp);

                        return 0;
                    }

                    inited = TRUE;

                    break;
                }

                pif = pif->ln_Succ;
            }
            pc = pc->ln_Succ;
        }
        psdUnlockDevice(pd);
    }
    DBG(KPrintF("sanei_usb_init device loop end\n"));

    psdUnlockPBase();
    DeleteMsgPort(mp);
    return inited ? 1 : 0;
}

///
/// sanei_usb_find_devices()
// attach device when vendor and product ids are found

SANE_Status sanei_usb_find_devices(SANE_Int vendor, SANE_Int product,
                                   SANE_Status(*attach)(SANE_String_Const dev))
{
    SANE_Int dn = 0;

    while(devices[dn].devname && (dn < MAX_DEVICES))
    {
        if((devices[dn].vendor == vendor) && (devices[dn].product == product))
        {
            if(attach)
            {
                attach(devices[dn].devname);
            }
        }
        dn++;
    }

    return SANE_STATUS_GOOD;
}

///
/// sanei_usb_set_timeout()

void sanei_usb_set_timeout(int dn, SANE_Int timeout)
{
    struct PsdPipe *outep;
    struct PsdPipe *inep;
    /*
    outep = devices[0].psd_Bulk[OUTEP].ep_BulkPipe;
    inep =  devices[0].psd_Bulk[INEP].ep_BulkPipe;

    psdSetAttrs(
    	PGA_PIPE, outep,
    	PPA_NakTimeout, TRUE,
    	PPA_NakTimeoutTime, timeout, // powiedzmy ze dobrze :)
    TAG_END);
        INTEP
    psdSetAttrs(
    	PGA_PIPE, inep,
    	PPA_NakTimeout, TRUE,
    	PPA_NakTimeoutTime, timeout, // powiedzmy ze dobrze :)
    TAG_END);
    */
    if((dn >= MAX_DEVICES) || (dn < 0) || !devices[dn].open)
    {
        DBG(KPrintF("sanei_usb_set_timeout: dn >= MAX_DEVICES || dn < 0\n"));
        return;
    }
    DBG(KPrintF("sanei_usb_set_timeout %d\n", dn));


    psdSetAttrs(
        PGA_PIPE, devices[dn].psd_DefaultPipe,
        PPA_NakTimeout, TRUE,
        PPA_NakTimeoutTime, timeout, // powiedzmy ze dobrze :)
        TAG_END);


}

///
/// sanei_usb_control_msg()

SANE_Status sanei_usb_control_msg(SANE_Int dn, SANE_Int rtype, SANE_Int req,
                                  SANE_Int value, SANE_Int index, SANE_Int len, SANE_Byte * data)
{
    int ioerr;

    if(dn >= MAX_DEVICES || dn < 0)
    {
        DBG(KPrintF("sanei_usb_control_msg: dn >= MAX_DEVICES || dn < 0, dn=%d\n",  dn));
        return SANE_STATUS_INVAL;
    }

    DBG(KPrintF("sanei_usb_control_msg: rtype = 0x%02x, req = %d, value = %d, index = %d, len = %d\n", rtype, req, value, index, len));

    psdPipeSetup(devices[dn].psd_DefaultPipe, rtype, req, value, index);
    ioerr = psdDoPipe(devices[dn].psd_DefaultPipe, data, len);

    if(ioerr > 0)
    {
        return SANE_STATUS_INVAL;
    }

    return SANE_STATUS_GOOD;

}

///
/// sanei_usb_attach_matching_devices()

/* This logically belongs to sanei_config.c but not every backend that
   uses sanei_config() wants to depend on sanei_usb.
   syntax: usb vendorID productID
*/
void sanei_usb_attach_matching_devices(const char *name,
                                       SANE_Status(*attach)(const char *dev))
{
    char *vendor, *product;

    if(strncmp(name, "usb", 3) == 0)
    {
        SANE_Word vendorID = 0, productID = 0;

        name += 3;

        name = sanei_config_skip_whitespace(name);
        if(*name)
        {
            name = sanei_config_get_string(name, &vendor);
            if(vendor)
            {
                vendorID = strtol(vendor, 0, 0);
                free(vendor);
            }
            name = sanei_config_skip_whitespace(name);
        }

        name = sanei_config_skip_whitespace(name);
        if(*name)
        {
            name = sanei_config_get_string(name, &product);
            if(product)
            {
                productID = strtol(product, 0, 0);
                free(product);
            }
        }
        sanei_usb_find_devices(vendorID, productID, attach);
    }
    else
    {
        (*attach)(name);
    }
}

///


SANE_Status sanei_usb_reset(SANE_Int dn)
{
    return SANE_STATUS_GOOD;
}

SANE_Status sanei_usb_read_int(SANE_Int dn, SANE_Byte * buffer, size_t * size)
{
    LONG ioerr, len;

    if(!size)
    {
        DBG(KPrintF("sanei_usb_read_int: size == NULL\n"));
        return SANE_STATUS_INVAL;
    }

    if(dn >= MAX_DEVICES || dn < 0 || !devices[dn].open)
    {
        DBG(KPrintF("sanei_usb_read_int: dn >= device number || dn < 0\n"));
        return SANE_STATUS_INVAL;
    }
    else
    {
        struct PsdPipe *intep =  devices[dn].psd_Bulk[INTEP].ep_BulkPipe;
        if(intep)
        {
            DBG(KPrintF("sanei_usb_read_int: trying to read %lu bytes\n", (unsigned long) *size));


            ioerr = psdDoPipe(intep, buffer, *size);
            len = psdGetPipeActual(intep);
            if(!len)
            {
                ioerr = psdDoPipe(intep, buffer, *size);
                len = psdGetPipeActual(intep);
            }
            *size = len;
            if(len > 0)
            {
                return SANE_STATUS_GOOD;
            }
            else if(!ioerr)
            {
                return SANE_STATUS_EOF;
            }

            DBG(KPrintF("sanei_usb_read_int() %s (ioerr %ld)\n", psdNumToStr(NTS_IOERR, ioerr, "Unknown Error"), ioerr));
            *size = 0;
        }
        else
        {
            DBG(KPrintF("sanei_usb_read_int: no interrupt pipe\n"));
        }
    }
    return SANE_STATUS_IO_ERROR;
}


int sanei_usb_exit(void)
{
    /* check we have really some work to do */
    if(inited == 0)
    {
        DBG(KPrintF("%s: sanei_usb in not initialized!\n", __func__));
        return 0;
    }

    inited = 0;
    return 1;
}

#if 0
SANE_Status sanei_usb_read_int(SANE_Int dn, SANE_Byte * buffer, size_t * size)
{
    ssize_t read_size = 0;
#if defined(HAVE_LIBUSB) || defined(HAVE_LIBUSB_1_0)
    SANE_Bool stalled = SANE_FALSE;
#endif

    if(!size)
    {
        DBG(1, "sanei_usb_read_int: size == NULL\n");
        return SANE_STATUS_INVAL;
    }

    if(dn >= device_number || dn < 0)
    {
        DBG(1, "sanei_usb_read_int: dn >= device number || dn < 0\n");
        return SANE_STATUS_INVAL;
    }

    DBG(5, "sanei_usb_read_int: trying to read %lu bytes\n", (unsigned long) *size);
    if(devices[dn].method == sanei_usb_method_scanner_driver)
    {
        DBG(1, "sanei_usb_read_int: access method %d not implemented\n",    devices[dn].method);
        return SANE_STATUS_INVAL;
    }
    else if(devices[dn].method == sanei_usb_method_libusb)
#ifdef HAVE_LIBUSB
    {
        if(devices[dn].int_in_ep)
        {
            read_size = usb_interrupt_read(devices[dn].libusb_handle,
                                           devices[dn].int_in_ep,
                                           (char *) buffer, (int) * size,
                                           libusb_timeout);

            if(read_size < 0)
                DBG(1, "sanei_usb_read_int: read failed: %s\n",
                    strerror(errno));

            stalled = (read_size == -EPIPE);
        }
        else
        {
            DBG(1, "sanei_usb_read_int: can't read without an int "
                "endpoint\n");
            return SANE_STATUS_INVAL;
        }
    }
#elif defined(HAVE_LIBUSB_1_0)
    {
        if(devices[dn].int_in_ep)
        {
            int ret;
            int trans_bytes;
            ret = libusb_interrupt_transfer(devices[dn].lu_handle,
                                            devices[dn].int_in_ep,
                                            buffer, (int) * size,
                                            &trans_bytes, libusb_timeout);

            if(ret < 0)
            {
                read_size = -1;
            }
            else
            {
                read_size = trans_bytes;
            }

            stalled = (ret == LIBUSB_ERROR_PIPE);
        }
        else
        {
            DBG(1, "sanei_usb_read_int: can't read without an int "
                "endpoint\n");
            return SANE_STATUS_INVAL;
        }
    }
#else /* not HAVE_LIBUSB && not HAVE_LIBUSB_1_0 */
    {
        DBG(1, "sanei_usb_read_int: libusb support missing\n");
        return SANE_STATUS_UNSUPPORTED;
    }
#endif /* not HAVE_LIBUSB && not HAVE_LIBUSB_1_0 */
    else if(devices[dn].method == sanei_usb_method_usbcalls)
    {
#ifdef HAVE_USBCALLS
        int rc;
        USHORT usNumBytes = *size;
        DBG(5, "Entered usbcalls UsbIrqStart with dn = %d\n", dn);
        DBG(5, "Entered usbcalls UsbIrqStart with dh = %p\n", dh);
        DBG(5, "Entered usbcalls UsbIrqStart with int_in_ep = 0x%02x\n", devices[dn].int_in_ep);
        DBG(5, "Entered usbcalls UsbIrqStart with interface_nr = %d\n", devices[dn].interface_nr);
        DBG(5, "Entered usbcalls UsbIrqStart with bytes to read = %u\n", usNumBytes);

        if(devices[dn].int_in_ep)
        {
            rc = UsbIrqStart(dh, devices[dn].int_in_ep, devices[dn].interface_nr,
                             usNumBytes, (char *) buffer, pUsbIrqStartHev);
            DBG(5, "rc of UsbIrqStart = %d\n", rc);
        }
        else
        {
            DBG(1, "sanei_usb_read_int: can't read without an int "
                "endpoint\n");
            return SANE_STATUS_INVAL;
        }
        if(rc)
        {
            return SANE_STATUS_INVAL;
        }
        read_size += usNumBytes;
#else
        DBG(1, "sanei_usb_read_int: usbcalls support missing\n");
        return SANE_STATUS_UNSUPPORTED;
#endif /* HAVE_USBCALLS */
    }
    else
    {
        DBG(1, "sanei_usb_read_int: access method %d not implemented\n",
            devices[dn].method);
        return SANE_STATUS_INVAL;
    }

    if(read_size < 0)
    {
#ifdef HAVE_LIBUSB
        if(devices[dn].method == sanei_usb_method_libusb)
            if(stalled)
            {
                usb_clear_halt(devices[dn].libusb_handle, devices[dn].int_in_ep);
            }
#elif defined(HAVE_LIBUSB_1_0)
        if(devices[dn].method == sanei_usb_method_libusb)
            if(stalled)
            {
                libusb_clear_halt(devices[dn].lu_handle, devices[dn].int_in_ep);
            }
#endif
        *size = 0;
        return SANE_STATUS_IO_ERROR;
    }
    if(read_size == 0)
    {
        DBG(3, "sanei_usb_read_int: read returned EOF\n");
        *size = 0;
        return SANE_STATUS_EOF;
    }
    DBG(5, "sanei_usb_read_int: wanted %lu bytes, got %ld bytes\n",
        (unsigned long) *size, (unsigned long) read_size);
    *size = read_size;
    if(debug_level > 10)
    {
        print_buffer(buffer, read_size);
    }

    return SANE_STATUS_GOOD;
}

#endif
