//
// Epson driver
//

#include <sane/sane.h>
#include <sane/saneopts.h>

#include <math.h>
#include <proto/exec.h>
#include <exec/types.h>
#include <string.h>
#include <devices/scanner.h>
#include "SaneInterface.h"

static int modeDepth[] = {8,24,1,1};
static int halfToneActive[] = {TRUE,FALSE,FALSE};
static int brightnessValues[] = {60,40,20,0,-20,-40,-60};

static SANE_Option_Descriptor* getGammaDescr(SANE_Handle handle,SANE_Int option);
static SANE_Status getGamma(SANE_Handle handle,SANE_Int option,void* val,ULONG* flags);
static SANE_Status setGamma(SANE_Handle handle,SANE_Int option,void* val,ULONG* flags);
static int saneNum(SANE_Handle handle,char* str);
int customGammaNum;
	   char a[]="Default";
	char b[]="User defined";
struct MicrotekInfo
{
  OptionId          ac_id;
  struct OptionInfo ac_info;
} info[] =
  {
    {ID_SCANMODE,       {0,SANE_NAME_SCAN_MODE,        modeDepth,NULL,NULL,NULL}},
    {ID_TL_X,           {0,SANE_NAME_SCAN_TL_X,             NULL,NULL,NULL,NULL}},
    {ID_TL_Y,           {0,SANE_NAME_SCAN_TL_Y,             NULL,NULL,NULL,NULL}},
    {ID_BR_X,           {0,SANE_NAME_SCAN_BR_X,             NULL,NULL,NULL,NULL}},
    {ID_BR_Y,           {0,SANE_NAME_SCAN_BR_Y,             NULL,NULL,NULL,NULL}},
    {ID_RESOLUTION,     {0,SANE_NAME_SCAN_RESOLUTION,       NULL,NULL,NULL,NULL}},
	{ID_BRIGHTNESS,     {0,SANE_NAME_BRIGHTNESS, NULL, NULL,NULL,NULL}},
    {ID_THRESHOLD,      {0,SANE_NAME_THRESHOLD,   NULL,NULL,NULL, NULL}},
    {ID_PREVIEW,        {0,SANE_NAME_PREVIEW,             NULL,NULL,NULL,NULL}},
	{ID_SCAN_SOURCE,    {0,SANE_NAME_SCAN_SOURCE,           NULL,NULL,NULL,NULL}},
    {ID_NONE,           {0,NULL,                            NULL,NULL,NULL,NULL}}
};

struct OptionInfo* getOption(SANE_Handle handle,OptionId id)
{
  int i = 0;

  while( (info[i].ac_id != ID_NONE) && (info[i].ac_id != id) )
    i++;
    
  if( info[i].ac_id == id )
    return &info[i].ac_info;
  else
    return NULL;
}



SANE_Status saneOpen(SANE_String_Const devicename,SANE_Handle* handle, char* firmware)
{
  SANE_Status stat;

  stat = sane_open(devicename,handle);
  return stat;
}


// Get SANE option number
//
//   input:  str = SANE option identifier
//
//   return: the SANE option number (0 if not supported)
//
static int saneNum(SANE_Handle handle,char* str)
{
  if( str == NULL )
    return -1;
  else
  {
    const SANE_Option_Descriptor *sod;
    int   i;

    i = 1;
    while( (sod = sane_get_option_descriptor(handle,i)) && ((sod->name==0) || strcmp(str,sod->name)) )
      i++;

    if( sod )
      return i;
    else
      return 0;
  }
}


